import unittest
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from threading import Event
import time
import tests_unit.testdatagenerator as testdatagenerator
from pelops import myconfigtools
from pelops.pubsub import mymqttclient
from pelops.logging import mylogger
from alcathous.datapoint import DataPoint
from alcathous.algorithms.average import Average
from alcathous.algorithms.weightedaverage import WeightedAverage
from alcathous.algorithms.count import Count
from alcathous.algorithms.maximum import Maximum
from alcathous.algorithms.minimum import Minimum
from alcathous.nodatabehavior import NoDataBehavior


class TestDatapoint(unittest.TestCase):
    def _on_response_avg(self, value):
        self.last_response_avg = value.decode("utf-8")
        self.on_response_avg_sync.set()

    def _wait_for_response_avg(self, message):
        if not self.on_response_avg_sync.wait(1):
            self.fail("Timeout waiting on response_avg for message '{}'.".format(message))
        self.on_response_avg_sync.clear()

    def _on_response_wavg(self, value):
        self.last_response_wavg = value.decode("utf-8")
        self.on_response_wavg_sync.set()

    def _wait_for_response_wavg(self, message):
        if not self.on_response_wavg_sync.wait(1):
            self.fail("Timeout waiting on response_wavg for message '{}'.".format(message))
        self.on_response_wavg_sync.clear()

    def _on_response_count(self, value):
        self.last_response_count = value.decode("utf-8")
        self.on_response_count_sync.set()

    def _wait_for_response_count(self, message):
        if not self.on_response_count_sync.wait(1):
            self.fail("Timeout waiting on response_count for message '{}'.".format(message))
        self.on_response_count_sync.clear()

    def _on_response_min(self, value):
        self.last_response_min = value.decode("utf-8")
        self.on_response_min_sync.set()

    def _wait_for_response_min(self, message):
        if not self.on_response_min_sync.wait(1):
            self.fail("Timeout waiting on response_min for message '{}'.".format(message))
        self.on_response_min_sync.clear()

    def _on_response_max(self, value):
        self.last_response_max = value.decode("utf-8")
        self.on_response_max_sync.set()

    def _wait_for_response_max(self, message):
        if not self.on_response_max_sync.wait(1):
            self.fail("Timeout waiting on response_max for message '{}'.".format(message))
        self.on_response_max_sync.clear()

    @classmethod
    def setUpClass(cls):
        cls.config = myconfigtools.read_config(
            os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))+"/tests_unit/config.yaml")
        cls.logger = mylogger.create_logger(cls.config["logger"], __name__)
        cls.logger.info("TestDatapoint - start")

    @classmethod
    def tearDownClass(cls):
        cls.logger.info("TestDatapoint - stop")

    def setUp(self):
        self.mqtt_client = mymqttclient.MyMQTTClient(self.config["pubsub"], self.logger, True)
        self.mqtt_client.connect()
        self.end_time, self.data_set = testdatagenerator.generate_test_data(number_entries=100, time_interval=30)

        self.on_response_avg_sync = Event()
        self.on_response_avg_sync.clear()
        self.last_response_avg = None
        self.on_response_wavg_sync = Event()
        self.on_response_wavg_sync.clear()
        self.last_response_wavg = None
        self.on_response_count_sync = Event()
        self.on_response_count_sync.clear()
        self.last_response_count = None
        self.on_response_min_sync = Event()
        self.on_response_min_sync.clear()
        self.last_response_min = None
        self.on_response_max_sync = Event()
        self.on_response_max_sync.clear()
        self.last_response_max = None

        self.config_methods = {}
        for m in self.config["data-preparation"]["methods"]:
            key = m["name"]
            if key in self.config_methods:
                raise ValueError("method name must be unique ({}).".format(key))
            self.config_methods[key] = m

        try:
            self.config_datapoint = self.config["data-preparation"]["datapoints"][0]
            self.no_data_behavior = NoDataBehavior.get_enum(self.config["data-preparation"]["no_data_behavior"])
        except KeyError:
            raise ValueError("config malformed.")

    def tearDown(self):
        self.mqtt_client.disconnect()

    def test_dp0_init(self):
        self.logger.info("TestDatapoint.test_dp_init ---------------------------------------------")
        dp = DataPoint(self.config_datapoint, self.config_methods, self.mqtt_client, self.logger,
                       self.no_data_behavior)
        self.assertEqual(dp._topic_sub, self.config_datapoint["topic-sub"])
        self.assertEqual(dp._topic_pub_prefix, self.config_datapoint["topic-pub-prefix"])
        self.assertEqual(dp._zero_is_valid, self.config_datapoint["zero_is_valid"])
        self.assertEqual(dp._max_time_window, 300)
        self.assertEqual(len(dp.methods), 5)
        method_types = []
        for m in dp.methods:
            method_types.append(type(m))
        self.assertEqual(len(method_types), 5)
        self.assertIn(Average, method_types)
        self.assertIn(WeightedAverage, method_types)
        self.assertIn(Count, method_types)
        self.assertIn(Maximum, method_types)
        self.assertIn(Minimum, method_types)
        self.assertIsNotNone(dp._data_set)
        self.assertEqual(len(dp._data_set), 0)
        self.assertIsNotNone(dp._lock_data_set)
        self.assertIsNotNone(dp._logger)
        self.assertIsNotNone(dp._pubsub_client)

    def test_dp1_purge(self):
        self.logger.info("TestDatapoint.test_dp_purge ---------------------------------------------")
        dp = DataPoint(self.config_datapoint, self.config_methods, self.mqtt_client, self.logger,
                       self.no_data_behavior)
        dp._data_set = self.data_set
        old_size = len(dp._data_set)
        dp.purge_old_values(self.end_time)
        new_size = len(dp._data_set)
        self.assertLess(new_size, old_size)
        self.assertEqual(new_size, 11)

    def test_dp2_send_data(self):
        self.logger.info("TestDatapoint.test_dp_send_data ---------------------------------------------")
        dp = DataPoint(self.config_datapoint, self.config_methods, self.mqtt_client, self.logger,
                       self.no_data_behavior)
        dp.start()
        for v in self.data_set.values():
            self.mqtt_client.publish(dp._topic_sub, v)
            time.sleep(0.025)
        time.sleep(0.5)
        self.assertEqual(len(dp._data_set), 100)
        dp.stop()

    def test_dp3_algorithms(self):
        self.logger.info("TestDatapoint.test_dp_algorithms ---------------------------------------------")
        dp = DataPoint(self.config_datapoint, self.config_methods, self.mqtt_client, self.logger,
                       self.no_data_behavior)
        for k,v in self.data_set.items():
            dp._data_set[k] = v
        dp.start()
        self.mqtt_client.subscribe(dp._topic_pub_prefix+"avg_5min", self._on_response_avg)
        self.mqtt_client.subscribe(dp._topic_pub_prefix+"wavg_5min", self._on_response_wavg)
        self.mqtt_client.subscribe(dp._topic_pub_prefix+"count_2min", self._on_response_count)
        self.mqtt_client.subscribe(dp._topic_pub_prefix+"min_3min", self._on_response_min)
        self.mqtt_client.subscribe(dp._topic_pub_prefix+"max_3min", self._on_response_max)
        for m in dp.methods:
            m.process(self.end_time)
        self._wait_for_response_avg("avg_5min")
        self._wait_for_response_wavg("wavg_5min")
        self._wait_for_response_count("count_2min")
        self._wait_for_response_min("min_3min")
        self._wait_for_response_max("max_3min")
        self.assertEqual(float(self.last_response_avg), 5)
        self.assertEqual(float(self.last_response_wavg), 3)
        self.assertEqual(int(self.last_response_count), 5)
        self.assertEqual(float(self.last_response_min), 0)
        self.assertEqual(float(self.last_response_max), 6)
        dp.stop()


if __name__ == '__main__':
    unittest.main()



from alcathous.algorithms.abstractalgorithm import AbstractAlgorithm


class WeightedAverage(AbstractAlgorithm):
    """
    The weighted average of all valid data points within the time window. The weight is the inverse time difference
    to the time_to time stamp.
    """

    def _process(self, time_from, time_to):
        self._logger.info("{} - weighted average for '{} s' to '{} s'.".format(self._topic, time_from, time_to))
        sum_weighted_values = 0
        weighted_count = 0
        with self._lock_data_set:
            self._logger.info("{} - acquired lock".format(self._topic))
            for timestamp, value in reversed(self._data_set.items()):
                if time_from > timestamp:
                    self._logger.info("{} - timestamp '{} s'; value '{}'.".format(self._topic, timestamp, value))
                    break
                elif timestamp <= time_to:
                    weight = timestamp - time_from
                    sum_weighted_values = sum_weighted_values + value * weight
                    weighted_count = weighted_count + weight
        self._logger.info("{} - weighted-sum: {}; weighted-count: {}.".
                  format(self._topic, sum_weighted_values, weighted_count))
        try:
            wavg = sum_weighted_values / weighted_count
        except ZeroDivisionError:
            raise ValueError
        return wavg

    def execution_points_estimation(self):
        return int(self._time_window * 1.1)



import unittest
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from threading import Lock, Event
import tests_unit.testdatagenerator as testdatagenerator
from pelops import myconfigtools
from pelops.pubsub import mymqttclient
from pelops.logging import mylogger
from alcathous.algorithms.average import Average
from alcathous.algorithms.weightedaverage import WeightedAverage
from alcathous.algorithms.count import Count
from alcathous.algorithms.maximum import Maximum
from alcathous.algorithms.minimum import Minimum
from alcathous.nodatabehavior import NoDataBehavior


class Tools:
    def _on_response(self, value):
        self.last_response = value.decode("utf-8")
        self.on_response_sync.set()

    def _wait_for_response(self, message):
        if not self.on_response_sync.wait(1):
            self.fail("Timeout waiting on below response for message '{}'.".format(message))
        self.on_response_sync.clear()

    def _setUp(self):
        self.mqtt_client = mymqttclient.MyMQTTClient(self.config["pubsub"], self.logger, True)
        self.mqtt_client.connect()
        self.lock_data_set = Lock()
        self.end_time, self.data_set = testdatagenerator.generate_test_data(number_entries=100, time_interval=30)
        self.on_response_sync = Event()
        self.on_response_sync.clear()
        self.last_response = None
        self.pos = 0
        try:
            for m in self.config["data-preparation"]["methods"]:
                if m["name"] == self.name:
                    self.config_alg = m
            self.pub_prefix = self.config["data-preparation"]["datapoints"][self.pos]["topic-pub-prefix"]
            self.pub_suffix = self.config_alg["topic-pub-suffix"]
            self.no_data_behavior = NoDataBehavior.get_enum(self.config["data-preparation"]["no_data_behavior"])
        except KeyError:
            raise ValueError("config malformed.")

    def _tearDown(self):
        self.mqtt_client.disconnect()


class TestAlgorithmAvg(unittest.TestCase, Tools):
    @classmethod
    def setUpClass(cls):
        cls.config = myconfigtools.read_config(
            os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))+"/tests_unit/config.yaml")
        cls.logger = mylogger.create_logger(cls.config["logger"], __name__)
        cls.logger.info("TestAlgorithmAvg - start")

    @classmethod
    def tearDownClass(cls):
        cls.logger.info("TestAlgorithmAvg - stop")

    def setUp(self):
        self.name = "avg_5min"
        self._setUp()

    def tearDown(self):
        self._tearDown()

    def test_avg0_init(self):
        alg = Average(self.name, self.config_alg, self.logger, self.data_set, self.lock_data_set, self.pub_prefix,
                      self.mqtt_client, self.no_data_behavior)
        self.assertEqual(alg.name, self.name)
        self.assertEqual(alg._time_window, 300)
        self.assertEqual(len(alg._data_set), 100)
        self.assertEqual(alg._no_data_behavior, NoDataBehavior.LAST_VALID)
        self.assertEqual(alg._last_valid, None)
        self.assertEqual(alg._topic, self.pub_prefix + self.pub_suffix)
        self.assertIsNotNone(alg._pubsub_client)
        self.assertIsNotNone(alg._logger)
        self.assertEqual(alg._config, self.config_alg)

    def test_avg1_execution_points_estimation(self):
        alg = Average(self.name, self.config_alg, self.logger, self.data_set, self.lock_data_set, self.pub_prefix,
                      self.mqtt_client, self.no_data_behavior)
        self.assertEqual(alg.execution_points_estimation(), 300)

    def test_avg2_process(self):
        alg = Average(self.name, self.config_alg, self.logger, self.data_set, self.lock_data_set, self.pub_prefix,
                      self.mqtt_client, self.no_data_behavior)
        self.mqtt_client.subscribe(alg._topic, self._on_response)
        alg.process(self.end_time)
        self._wait_for_response("process average")
        self.assertIsNotNone(self.last_response)
        self.assertEqual(alg._last_valid, 5.0)
        self.assertEqual(float(self.last_response), 5.0)


class TestAlgorithmWAvg(unittest.TestCase, Tools):
    @classmethod
    def setUpClass(cls):
        cls.config = myconfigtools.read_config(
            os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))+"/tests_unit/config.yaml")
        cls.logger = mylogger.create_logger(cls.config["logger"], __name__)
        cls.logger.info("TestAlgorithmWAvg - start")

    @classmethod
    def tearDownClass(cls):
        cls.logger.info("TestAlgorithmWAvg - stop")

    def setUp(self):
        self.name = "wavg_5min"
        self._setUp()

    def tearDown(self):
        self._tearDown()

    def test_wavg0_init(self):
        alg = WeightedAverage(self.name, self.config_alg, self.logger, self.data_set, self.lock_data_set,
                              self.pub_prefix, self.mqtt_client, self.no_data_behavior)
        self.assertEqual(alg.name, self.name)
        self.assertEqual(alg._time_window, 300)
        self.assertEqual(len(alg._data_set), 100)
        self.assertEqual(alg._no_data_behavior, NoDataBehavior.LAST_VALID)
        self.assertEqual(alg._last_valid, None)
        self.assertEqual(alg._topic, self.pub_prefix + self.pub_suffix)
        self.assertIsNotNone(alg._pubsub_client)
        self.assertIsNotNone(alg._logger)
        self.assertEqual(alg._config, self.config_alg)

    def test_wavg1_execution_points_estimation(self):
        alg = WeightedAverage(self.name, self.config_alg, self.logger, self.data_set, self.lock_data_set,
                              self.pub_prefix, self.mqtt_client, self.no_data_behavior)
        self.assertEqual(alg.execution_points_estimation(), 330)

    def test_wavg2_process(self):
        alg = WeightedAverage(self.name, self.config_alg, self.logger, self.data_set, self.lock_data_set,
                              self.pub_prefix, self.mqtt_client, self.no_data_behavior)
        self.mqtt_client.subscribe(alg._topic, self._on_response)
        alg.process(self.end_time)
        self._wait_for_response("process weighted average")
        self.assertIsNotNone(self.last_response)
        self.assertEqual(alg._last_valid, 3.0)
        self.assertEqual(float(self.last_response), 3.0)


class TestAlgorithmCount(unittest.TestCase, Tools):
    @classmethod
    def setUpClass(cls):
        cls.config = myconfigtools.read_config(
            os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))+"/tests_unit/config.yaml")
        cls.logger = mylogger.create_logger(cls.config["logger"], __name__)
        cls.logger.info("TestAlgorithmCount - start")

    @classmethod
    def tearDownClass(cls):
        cls.logger.info("TestAlgorithmCount - stop")

    def setUp(self):
        self.name = "count_2min"
        self._setUp()

    def tearDown(self):
        self._tearDown()

    def test_count0_init(self):
        alg = Count(self.name, self.config_alg, self.logger, self.data_set, self.lock_data_set,
                    self.pub_prefix, self.mqtt_client, self.no_data_behavior)
        self.assertEqual(alg.name, self.name)
        self.assertEqual(alg._time_window, 120)
        self.assertEqual(len(alg._data_set), 100)
        self.assertEqual(alg._no_data_behavior, NoDataBehavior.LAST_VALID)
        self.assertEqual(alg._last_valid, None)
        self.assertEqual(alg._topic, self.pub_prefix + self.pub_suffix)
        self.assertIsNotNone(alg._pubsub_client)
        self.assertIsNotNone(alg._logger)
        self.assertEqual(alg._config, self.config_alg)

    def test_count1_execution_points_estimation(self):
        alg = Count(self.name, self.config_alg, self.logger, self.data_set, self.lock_data_set,
                    self.pub_prefix, self.mqtt_client, self.no_data_behavior)
        self.assertEqual(alg.execution_points_estimation(), 108)

    def test_count2_process(self):
        alg = Count(self.name, self.config_alg, self.logger, self.data_set, self.lock_data_set,
                    self.pub_prefix, self.mqtt_client, self.no_data_behavior)
        self.mqtt_client.subscribe(alg._topic, self._on_response)
        alg.process(self.end_time)
        self._wait_for_response("process count")
        self.assertIsNotNone(self.last_response)
        self.assertEqual(alg._last_valid, 5)
        self.assertEqual(int(self.last_response), 5)


class TestAlgorithmMin(unittest.TestCase, Tools):
    @classmethod
    def setUpClass(cls):
        cls.config = myconfigtools.read_config(
            os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))+"/tests_unit/config.yaml")
        cls.logger = mylogger.create_logger(cls.config["logger"], __name__)
        cls.logger.info("TestAlgorithmMin - start")

    @classmethod
    def tearDownClass(cls):
        cls.logger.info("TestAlgorithmMin - stop")

    def setUp(self):
        self.name = "min_3min"
        self._setUp()

    def tearDown(self):
        self._tearDown()

    def test_min0_init(self):
        alg = Minimum(self.name, self.config_alg, self.logger, self.data_set, self.lock_data_set,
                      self.pub_prefix, self.mqtt_client, self.no_data_behavior)
        self.assertEqual(alg.name, self.name)
        self.assertEqual(alg._time_window, 180)
        self.assertEqual(len(alg._data_set), 100)
        self.assertEqual(alg._no_data_behavior, NoDataBehavior.LAST_VALID)
        self.assertEqual(alg._last_valid, None)
        self.assertEqual(alg._topic, self.pub_prefix + self.pub_suffix)
        self.assertIsNotNone(alg._pubsub_client)
        self.assertIsNotNone(alg._logger)
        self.assertEqual(alg._config, self.config_alg)

    def test_min1_execution_points_estimation(self):
        alg = Minimum(self.name, self.config_alg, self.logger, self.data_set, self.lock_data_set,
                      self.pub_prefix, self.mqtt_client, self.no_data_behavior)
        self.assertEqual(alg.execution_points_estimation(), 180)

    def test_min2_process(self):
        alg = Minimum(self.name, self.config_alg, self.logger, self.data_set, self.lock_data_set,
                      self.pub_prefix, self.mqtt_client, self.no_data_behavior)
        self.mqtt_client.subscribe(alg._topic, self._on_response)
        alg.process(self.end_time)
        self._wait_for_response("process min")
        self.assertIsNotNone(self.last_response)
        self.assertEqual(alg._last_valid, 0)
        self.assertEqual(float(self.last_response), 0)


class TestAlgorithmMax(unittest.TestCase, Tools):
    @classmethod
    def setUpClass(cls):
        cls.config = myconfigtools.read_config(
            os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))+"/tests_unit/config.yaml")
        cls.logger = mylogger.create_logger(cls.config["logger"], __name__)
        cls.logger.info("TestAlgorithmMax - start")

    @classmethod
    def tearDownClass(cls):
        cls.logger.info("TestAlgorithmMax - stop")

    def setUp(self):
        self.name = "max_3min"
        self._setUp()

    def tearDown(self):
        self._tearDown()

    def test_max0_init(self):
        alg = Maximum(self.name, self.config_alg, self.logger, self.data_set, self.lock_data_set,
                      self.pub_prefix, self.mqtt_client, self.no_data_behavior)
        self.assertEqual(alg.name, self.name)
        self.assertEqual(alg._time_window, 180)
        self.assertEqual(len(alg._data_set), 100)
        self.assertEqual(alg._no_data_behavior, NoDataBehavior.LAST_VALID)
        self.assertEqual(alg._last_valid, None)
        self.assertEqual(alg._topic, self.pub_prefix + self.pub_suffix)
        self.assertIsNotNone(alg._pubsub_client)
        self.assertIsNotNone(alg._logger)
        self.assertEqual(alg._config, self.config_alg)

    def test_max1_execution_points_estimation(self):
        alg = Maximum(self.name, self.config_alg, self.logger, self.data_set, self.lock_data_set,
                      self.pub_prefix, self.mqtt_client, self.no_data_behavior)
        self.assertEqual(alg.execution_points_estimation(), 180)

    def test_max2_process(self):
        alg = Maximum(self.name, self.config_alg, self.logger, self.data_set, self.lock_data_set,
                      self.pub_prefix, self.mqtt_client, self.no_data_behavior)
        self.mqtt_client.subscribe(alg._topic, self._on_response)
        alg.process(self.end_time)
        self._wait_for_response("process max")
        self.assertIsNotNone(self.last_response)
        self.assertEqual(alg._last_valid, 6)
        self.assertEqual(float(self.last_response), 6)


if __name__ == '__main__':
    unittest.main()



from pelops.abstractmicroservice import AbstractMicroservice
import time
import threading
from pelops.mythreading import LoggerThread
from alcathous.datapoint import DataPoint
from alcathous.nodatabehavior import NoDataBehavior
from alcathous.worker import Worker
import alcathous.schema as schema
from alcathous import version


class DataPointManager(AbstractMicroservice):
    """
    Manage the data points by creating/start/stopping them and invocation of the process methods of the algorithms
    regularly.

    Config yaml:
    mqtt:
        mqtt-address: localhost
        mqtt-port: 1883
        credentials-file: ~/credentials.yaml
        log-level: INFO

    logger:
        log-level: DEBUG
        log-file: alcathous.log

    data-preparation:  # alcathous root node
        no_data_behavior: last_valid  # mute, last_valid, empty_message
        update_cycle: 30  # new values published each ... seconds
        number_worker: 2  # how many worker threads should be spawned to process task queue

        methods:
            avg_5min:
                topic-pub-suffix: avg_5min
                algorithm: avg  # avg - average, wavg - weighted average, count, min, max
                time_window: 5  # use the values from the last ... minutes

            wavg_5min:
                topic-pub-suffix: wavg_5min
                algorithm: wavg  # avg - average, wavg - weighted average, count, min, max
                time_window: 5  # use the values from the last ... minutes

        datapoints:
            - topic-sub: /test/0/raw
              topic-pub-prefix: /test/0/aggregated/
              zero_is_valid: False  # 0 is valid or rejected
              methods: wavg_5min, avg_5min
    """
    _version = version

    _data_points = None  # list of datapoint instances
    _processes = None  # list with references to all process functions of the algorithms of the datapoints
    _purges = None  # list with references to all purge functions of the datapoints
    _no_data_behavior = None  # defines how the algorithm should react if no data is available.
    _update_cycle = None  # call all processes every n seconds
    _worker = None  # instance of the worker manager
    _do_loop_thread = None  # contains the loop method thread

    def __init__(self, config, pubsub_client=None, logger=None, stdout_log_level=None, no_gui=None):
        """
        Constructor.

        :param config: config yaml structure
        :param pubsub_client: instance of an mymqttclient (optional)
        :param logger: instance of a logger (optional)
        :param no_gui: if False create and control a ui instance
        :param stdout_log_level: if set, a logging handler with target sys.stdout will be added
        """

        AbstractMicroservice.__init__(self, config, "data-preparation", pubsub_client=pubsub_client, logger=logger,
                                      stdout_log_level=stdout_log_level, no_gui=no_gui)

        self._purges = []
        self._processes = []
        self._data_points = []

        self._update_cycle = int(self._config["update_cycle"])
        self._no_data_behavior = NoDataBehavior.get_enum(str(self._config["no_data_behavior"]))

        self._worker = Worker(self._config["number_worker"], self._logger)
        self._do_loop_thread = LoggerThread(target=self._do_loop, name="datapointmanager", logger=self._logger)

        _config_methods = {}
        for m in self._config["methods"]:
            key = m["name"]
            if key in _config_methods:
                self._logger.error("DataPointManager - method name must be unique ({}).".format(key))
                raise ValueError("DataPointManager - method name must be unique ({}).".format(key))
            _config_methods[key] = m

        for config_data_point in self._config["datapoints"]:
            dp = DataPoint(config_data_point, _config_methods, self._pubsub_client, self._logger,
                           self._no_data_behavior)
            self._purges.append(dp.purge_old_values)
            for method in dp.methods:
                process = method.process
                cost = method.execution_points_estimation()
                self._logger.info("DataPointManager - adding process '{}' with cost '{}'.".
                                  format(process.__name__, cost))
                self._processes.append((process, cost))
            self._data_points.append(dp)

        self._processes.sort(key=lambda tup: tup[1], reverse=True)  # sort processes by their cost most expensive first

    @classmethod
    def _get_description(cls):
        return "This software subscribes to mqtt-topics that contain raw sensor data and publishes average values " \
               "for configurable time spans."

    def _loop_process(self):
        """
        Call all process and purge methods.
        """

        timestamp = time.time()
        self._logger.info("DataPointManager - started work for timestamp '{} s'.".format(timestamp))
        for p in self._processes:
            self._worker.queue.put((p[0], timestamp))
            self._logger.info("DataPointManager - waiting for worker to finish processing the algorithms.")
        self._worker.queue.join()
        for p in self._purges:
            self._worker.queue.put((p, timestamp))
        self._logger.info("DataPointManager - waiting for worker to purge outdated values.")
        self._worker.queue.join()

    def _do_loop(self):
        """
        Call _loop_process every n seconds.
        """
        self._logger.info("DataPointManager - start loop.")
        while not self._stop_service.is_set():
            start = time.time()
            self._loop_process()
            sleep_for = max(0, self._update_cycle - (time.time() - start))
            self._logger.info("DataPointManager - wait for '{} s'.".format(sleep_for))
            self._stop_service.wait(sleep_for)
        self._logger.info("DataPointManager - exited loop.")

    @classmethod
    def _get_schema(cls):
        """
        Get the sub schema to validate the yaml-config file against.

        :return: json-schema dict
        """
        return schema.get_schema()

    def _start(self):
        """
        Start all datapoints, worker and the process loop.
        """
        for dp in self._data_points:
            dp.start()
        self._worker.start()
        self._do_loop_thread.start()

    def _stop(self):
        """
        Stop all datapoints, worker and the process loop.
        """
        for dp in self._data_points:
            dp.stop()
        self._do_loop_thread.join()
        self._worker.stop()

    def runtime_information(self):
        return {}

    def config_information(self):
        return {}


def standalone():
    """Calls the static method DataPointManager.standalone()."""
    DataPointManager.standalone()


if __name__ == "__main__":
    DataPointManager.standalone()



from alcathous.nodatabehavior import NoDataBehavior
from pelops.logging import mylogger


class AbstractAlgorithm:
    """
    AbstractAlgorithm provides everything that is needed for an algorithm to perform its aggregation work. It selects
    to correct subset of the collected data.

    config yaml:
        topic-pub-suffix: avg_5min
        algorithm: avg  # avg - average, wavg - weighted average, count, min, max
        time_window: 5  # use the values from the last ... minutes
    """

    _data_set = None  # ordered dict containing values and timestamps. It is owned the parent datapoint class instance.
    _lock_data_set = None  # prevents datapoint class to change the dataset while an algorithm process it.
    _time_window = None  # time window in seconds. e.g. use the values from the last 300s and calculate their average.
    _no_data_behavior = None  # defines how the algorithm should react if no data is available.
    _last_valid = None  # stores the last valid result of the algorithm.
    _topic = None  # publish the result to this topic
    _pubsub_client = None  # instance of pelops.pubsub client
    name = None  # identifier for this algorithm
    _config = None  # stores the yaml-configuration for this algorithm
    _logger = None  # logger instance

    def __init__(self, name, config, parent_logger, data_set, lock_data_set, topic_pub_prefix, pubsub_client,
                 no_data_behavior):
        """
        Constructor

        :param name: Name of the instance of the algorithm
        :param config: yaml config structure
        :param parent_logger: logger instance from the parent. a child will be spawned
        :param data_set: the data the algortihm should work with
        :param lock_data_set: the algorithm works with the data set only if a lock has been acquired successfully
        :param topic_pub_prefix: prefix+name is the topic the result will be published to
        :param pubsub_client: instance of an mymqttclient
        :param no_data_behavior: defines how the algorithm should react if no data is available.
        """
        self._logger = mylogger.get_child(parent_logger, self.__class__.__name__)
        self.name = name
        self._config = config
        self._no_data_behavior = no_data_behavior
        self._data_set = data_set
        self._lock_data_set = lock_data_set

        self._logger.info("{}.{} - __init__".format(self.__class__.__name__, name))
        self._logger.debug("{}.{} - __init__ config: ".format(self.__class__.__name__, name, self._config))
        self._topic = topic_pub_prefix + self._config["topic-pub-suffix"]
        self._logger.info("{}.{} - publish to topic '{}'.".format(self.__class__.__name__, name, self._topic))

        self._pubsub_client = pubsub_client
        self._time_window = int(self._config["time_window"]) * 60
        if self._time_window <= 0:
            self._logger.error("Value for time_window must be a positive integer larger than 0. ('{}' not > 0)".
                               format(self._time_window))
            raise ValueError("Value for time_window must be a positive integer larger than 0. ('{}' not > 0)".
                             format(self._time_window))
        self._logger.info("{}.{} - time window {} s.".format(self.__class__.__name__, name, self._time_window))

    def process(self, timestamp):
        """
        Calls the specialiced method _process and publishes the result to _topic.

        :param timestamp: timestamp of the latest value to be used. valid timestamps are within timestamp and
        (timestamp - self._time_window)
        """
        time_from = timestamp - self._time_window
        time_to = timestamp
        try:
            value = self._process(time_from, time_to)
            self._last_valid = value
            self._pubsub_client.publish(self._topic, value)
        except ValueError:
            self._logger.info("{} - process/ValueError. performing no data behavior '{}'.".
                              format(self._topic, self._no_data_behavior))
            if self._no_data_behavior == NoDataBehavior.MUTE:
                pass
            elif self._no_data_behavior == NoDataBehavior.EMPTY_MESSAGE:
                self._pubsub_client.publish(self._topic, None)
            elif self._no_data_behavior == NoDataBehavior.LAST_VALID:
                self._pubsub_client.publish(self._topic, self._last_valid)
            else:
                self._logger.error("Don't know how to handle NoDataBehavior.{}.".format(self._no_data_behavior))
                raise NotImplementedError("Don't know how to handle NoDataBehavior.{}.".format(self._no_data_behavior))

    def _process(self, time_from, time_to):
        """
        This method must be implemented by each implementation of this abstract algorithm class. This is the core
        functionality - whatever condensation with the provided data should be done it must be implemented within
        this method.

        If no_data occurs (no/not enough valid values within given time window) then a ValueError exception should be raised.

        :param time_from: earliest entry from data_set
        :param time_to: latest entry from data_set
        :return: the result of the condensation
        :raises: ValueError if no_data occured.
        """
        raise NotImplementedError()

    def execution_points_estimation(self):
        """
        Each implementation of this abstract algorithm class must provide an estimation how much execution cost.

        :return: numeric value larger than 0
        """
        raise NotImplementedError()



import collections
from threading import Lock
import time
from alcathous.algorithms.algorithmfactory import AlgorithmFactory
from pelops.logging import mylogger


class DataPoint:
    """
    A DataPoint subscribes to a single topic and applies n methods/algorithms to the recorded data set.

    config yaml for the datapoint:
      - topic-sub: /test/0/raw
        topic-pub-prefix: /test/0/aggregated/
        zero_is_valid: False  # 0 is valid or rejected
        methods: wavg_5min, avg_5min

    config yaml for the methods:

    methods:
        avg_5min:
            topic-pub-suffix: avg_5min
            algorithm: avg  # avg - average, wavg - weighted average, count, min, max
            time_window: 5  # use the values from the last ... minutes

        wavg_5min:
            topic-pub-suffix: wavg_5min
            algorithm: wavg  # avg - average, wavg - weighted average, count, min, max
            time_window: 5  # use the values from the last ... minutes
    """

    methods = None  # list of instantiated methods / algorithms
    _topic_sub = None  # the values received through this topic subscription should be processed
    _topic_pub_prefix = None  # prefix for publishing the method results
    _zero_is_valid = None  # is 0 a valid value or should it be skipped
    _max_time_window = None  # each method has its own time window. this variable stores the maximum for this datapoint.
    _data_set = None  # ordered dict containing values and timestamps. It is owned the parent datapoint class instance.
    _lock_data_set = None  # prevents datapoint class to change the dataset while an algorithm process it.
    _logger = None  # logger instance
    _pubsub_client=None  # instance of pelops.mymqttclient

    def __init__(self, config_datapoint, config_methods, pubsub_client, parent_logger, no_data_behavior):
        """
        Constructor.

        :param config_datapoint: config yaml structure of this data point
        :param config_methods: config yaml structure for all methods
        :param parent_logger: logger instance from the parent. a child will be spawned
        :param pubsub_client: instance of an mymqttclient
        :param no_data_behavior: defines how the algorithm should react if no data is available.
        """

        self._logger = mylogger.get_child(parent_logger, self.__class__.__name__)
        self._logger.info("DataPoint.__init__ - initializing")
        self._logger.debug("DataPoint.__init__ - config_datapoint: {}".format(config_datapoint))
        self._logger.debug("DataPoint.__init__ - config_methods: {}".format(config_methods))
        self._pubsub_client = pubsub_client
        self._topic_sub = str(config_datapoint["topic-sub"])
        self._logger.info("DataPoint.__init__ - topic_sub: {}".format(self._topic_sub))
        self._topic_pub_prefix = str(config_datapoint["topic-pub-prefix"])
        self._logger.info("{} - publish to '{}#'".format(self._topic_sub, self._topic_pub_prefix))
        self._data_set = collections.OrderedDict()
        self._lock_data_set = Lock()
        self._max_time_window = 0
        self._zero_is_valid = bool(config_datapoint["zero_is_valid"])

        temp_methods = [x.strip() for x in config_datapoint["methods"].split(',')]
        self.methods = AlgorithmFactory.get_instances(temp_methods, config_methods, self._logger, self._data_set,
                                                      self._lock_data_set, self._topic_pub_prefix, self._pubsub_client,
                                                      no_data_behavior)
        for m in self.methods:
            if m._time_window > self._max_time_window:
                self._max_time_window = m._time_window

        self._logger.info("{} - max time window for purging data: {} s.".format(self._topic_sub, self._max_time_window))

    def _message_handler(self, value):
        """
        Message handler - to be registered to _topic_sub in pubsub_client. Stores the received value in _data_set with
        the current time stamp.

        :param value: Message content from incoming mqtt message.
        """

        if self._is_value_valid(value):
            with self._lock_data_set:
                timestamp = time.time()
                self._data_set[timestamp] = float(value)
                self._logger.info("{} - added {}@{}s".format(self._topic_sub, value, timestamp))
                self._logger.debug(self._data_set)

    def _is_value_valid(self, value):
        """
        Checks for simple rules: 'is None' and '==0' if _zero_is_valid is not set.

        :param value:
        :return: True/False
        """

        result = True
        if value is None:
            result = False
        elif value == 0 and not self._zero_is_valid:
            result = False

        if not result:
            self._logger.info("{} - value '{}' is not valid.".format(self._topic_sub, value))
        return result

    def purge_old_values(self, timestamp):
        """
        Remove all entries in _data_set that have a time stamp that is older than timestamp.

        :param timestamp: oldest time stamp to be kept.
        """

        min_time_stamp = timestamp - self._max_time_window
        self._logger.info("{} - purging values with timestamp < '{}'.".format(self._topic_sub, min_time_stamp))
        count = 0
        list_size = len(self._data_set)
        with self._lock_data_set:
            while len(self._data_set) and (next(iter(self._data_set.items())))[0] < min_time_stamp:
                self._logger.debug("{} - purge item '{}'.".format(self._topic_sub, (next(iter(self._data_set.items())))))
                count = count + 1
                self._data_set.popitem(False)
        self._logger.info("{} -  purged {}/{} items.".format(self._topic_sub, count, list_size))

    def start(self):
        """
        Start by subscribing to _topic_sub.
        """

        self._logger.info("Datapoint.start - subscribing to topic '{}'.".format(self._topic_sub))
        self._pubsub_client.subscribe(self._topic_sub, self._message_handler)

    def stop(self):
        """
        Stop by unsubscription from _topic_sub.
        """

        self._logger.info("Datapoint.stop - unsubscribing to topic '{}'.".format(self._topic_sub))
        self._pubsub_client.unsubscribe(self._topic_sub, self._message_handler)
